# -*- coding: utf-8 -*-

# Copyright 2010, 2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
from skoolparser import get_address, parse_int

BLOCKS = 'b'
BLOCK_TITLES = 't'
BLOCK_DESC = 'd'
REGISTERS = 'r'
BLOCK_COMMENTS = 'm'
SUBBLOCKS = 's'
COMMENTS = 'c'

def get_defb_length(op):
    return op.count(',') + 1

def get_defw_length(op):
    return 2 * (op.count(',') + 1)

def get_defm_length(op):
    start = op.index('"')
    return op.index('"', start + 1) - start - 1

def get_instruction_ctl(op):
    if op.startswith('DEFB'):
        return 'b'
    if op.startswith('DEFW'):
        return 'w'
    if op.startswith('DEFM'):
        return 't'
    if op.startswith('DEFS'):
        return 'z'
    return 'c'

class FakeInstruction:
    def __init__(self, address, comment):
        self.address = address
        self.comment = comment

class CtlWriter:
    def __init__(self, entries, elements='btdrmsc', write_hex=False):
        self.entries = entries
        self.elements = elements
        self.write_hex = write_hex
        self.out = sys.stdout

    def write(self, out=None):
        if out:
            self.out = out
        for entry in self.entries:
            self.write_entry(entry)

    def write_line(self, text):
        self.out.write('%s\n' % text)

    def write_entry(self, entry):
        address = self.addr_str(entry.address)
        if BLOCKS in self.elements:
            if BLOCK_TITLES in self.elements:
                self.write_line(('%s %s %s' % (entry.ctl, address, entry.description)).rstrip())
            else:
                self.write_line('%s %s' % (entry.ctl, address))
        if BLOCK_DESC in self.elements:
            if entry.details:
                self.write_line('D %s %s' % (address, entry.details))
        if REGISTERS in self.elements:
            for reg in entry.registers:
                self.write_line('R %s %s %s' % (address, reg.name, reg.contents))
        self.write_body(entry)
        if BLOCK_COMMENTS in self.elements and entry.end_comment:
            self.write_line('E %s %s' % (address, entry.end_comment))

    def write_body(self, entry):
        entry_ctl = 'b' if entry.ctl in 'gu' else entry.ctl
        first_instruction = entry.instructions[0]
        if entry_ctl == 'i' and not first_instruction.operation:
            # Don't write any sub-blocks for an empty 'i' entry
            return

        # Split the entry into sections separated by mid-routine comments
        sections = [('', [first_instruction])]
        for instruction in entry.instructions[1:]:
            mrc = instruction.get_mid_routine_comment()
            if mrc:
                sections.append((mrc, [instruction]))
            else:
                sections[-1][1].append(instruction)

        for k, (mrc, instructions) in enumerate(sections):
            if BLOCK_COMMENTS in self.elements and mrc:
                address = instructions[0].address
                self.write_line('D %s %s' % (self.addr_str(address), mrc))

            if SUBBLOCKS in self.elements:
                sub_blocks = self.get_sub_blocks(instructions)
                for j, (ctl, instructions) in enumerate(sub_blocks):
                    if ctl != 'M' or COMMENTS in self.elements:
                        length = None
                        offset = instructions[0].comment.rowspan + 1 if ctl == 'M' else 1
                        if j + offset < len(sub_blocks):
                            length = int(sub_blocks[j + offset][1][0].address) - int(instructions[0].address)
                        elif k + 1 < len(sections):
                            length = int(sections[k + 1][1][0].address) - int(instructions[0].address)
                        comment = ''
                        if instructions[0].comment and COMMENTS in self.elements:
                            comment = instructions[0].comment.text
                        if self.need_sub_block(ctl, entry_ctl, comment, j, instructions):
                            self.write_sub_block(ctl, entry_ctl, comment, j, instructions, length)

    def addr_str(self, address):
        return '$%04X' % int(address) if self.write_hex else str(address)

    def num_str(self, num):
        if not self.write_hex:
            return str(num)
        integer = int(num)
        if integer < 10:
            fmt = '%i'
        elif integer < 256:
            fmt = '$%02X'
        else:
            fmt = '$%04X'
        return fmt % integer

    def need_sub_block(self, ctl, entry_ctl, comment, index, instructions):
        # Return whether a ctl line needs to be written for a sub-block
        if ctl == entry_ctl and not comment:
            # Don't write a ctl line for a commentless sub-block of a code block
            if ctl == 'c':
                return False
            if index == 0 and len(instructions) == 1:
                # Don't write a ctl line for a DEF{B,M,W} statement if it's 1 byte
                # or 1 word, commentless, the first instruction in the section, the
                # sole instruction in its sub-block, and of the same type as the
                # containing block (i.e. 'b', 't' or 'w')
                if ctl == 'b' and get_defb_length(instructions[0].operation) == 1:
                    return False
                if ctl == 't' and get_defm_length(instructions[0].operation) == 1:
                    return False
                if ctl == 'w' and get_defw_length(instructions[0].operation) == 2:
                    return False
                # Don't write a ctl line for a DEFS statement if it's commentless,
                # the first instruction in the section, the sole instruction in its
                # sub-block, and inside a 'z' block
                if ctl == 'z':
                    return False
        return True

    def get_sub_blocks(self, instructions):
        # Split a block of instructions into sub-blocks by comment rowspan and/or
        # instruction type
        sub_blocks = []
        i = 0
        prev_ctl = ''
        while i < len(instructions):
            instruction = instructions[i]
            comment = instruction.comment
            ctl = get_instruction_ctl(instruction.operation)
            if comment and (comment.rowspan > 1 or comment.text):
                inst_ctls = set()
                for inst in instructions[i:i + comment.rowspan]:
                    inst_ctls.add(get_instruction_ctl(inst.operation))
                if len(inst_ctls) > 1:
                    # We've found a set of two or more instructions of various
                    # types with a single comment, so add a commented 'M' sub-block
                    # and commentless sub-blocks for the instructions
                    sub_blocks.append(('M', [FakeInstruction(instruction.address, instruction.comment)]))
                    instruction.comment = None
                    sub_blocks += self.get_sub_blocks(instructions[i:i + comment.rowspan])
                else:
                    # We've found a set of one or more instructions of the same
                    # type with a comment, so add a new sub-block
                    sub_blocks.append((ctl, instructions[i:i + comment.rowspan]))
                prev_ctl = ''
            elif ctl == prev_ctl:
                # This instruction is commentless and is of the same type as the
                # previous instruction (which was also commentless), so add it to
                # the current sub-block
                sub_blocks[-1][1].append(instruction)
            else:
                # This instruction is commentless but of a different type from the
                # previous instruction, so start a new sub-block
                sub_blocks.append((ctl, [instruction]))
                prev_ctl = ctl
            i += comment.rowspan if comment else 1
        return sub_blocks

    def get_lengths(self, stmt_lengths):
        # Find subsequences of indentical statement lengths and abbreviate them,
        # e.g. '16,16,16,8,8,4' -> '16*3,8*2,4'
        lengths = []
        prev = None
        for length in stmt_lengths:
            if length == prev:
                lengths[-1][1] += 1
            else:
                lengths.append([length, 1])
                prev = length
        length_params = []
        for length, mult in lengths:
            if mult == 1:
                length_params.append(self.num_str(length))
            else:
                length_params.append('%s*%s' % (self.num_str(length), self.num_str(mult)))
        return length_params

    def write_sub_block(self, ctl, entry_ctl, comment, index, instructions, length):
        sub_block_ctl = ' ' if ctl == entry_ctl else ctl.upper()
        address = self.addr_str(instructions[0].address)
        lengths = ''

        if ctl in 'btwz':
            # Find the byte lengths of each line in a 'B', 'T', 'W' or 'Z'
            # sub-block
            stmt_lengths = []
            for stmt in instructions:
                operation = stmt.operation
                if ctl == 'b':
                    stmt_lengths.append(get_defb_length(operation))
                elif ctl == 't':
                    stmt_lengths.append(get_defm_length(operation))
                elif ctl == 'w':
                    stmt_lengths.append(get_defw_length(operation))
                else:
                    stmt_lengths.append(parse_int(get_address(operation[4:])))
            num_bytes = sum(stmt_lengths)
            while len(stmt_lengths) > 1 and stmt_lengths[-1] == stmt_lengths[-2]:
                stmt_lengths.pop()
            # Don't write the statement lengths in a ctl line for a lone DEFB
            # statement of length 1, a lone DEFM statement of length 1, a lone
            # DEFW statement of length 2, or a lone DEFS statement
            if not (len(instructions) == 1 and ((ctl in 'bt' and num_bytes == 1) or (ctl == 'w' and num_bytes == 2) or ctl == 'z')):
                length = num_bytes
                lengths = ',' + ','.join(self.get_lengths(stmt_lengths))

        if length:
            self.write_line(('%s %s,%s%s %s' % (sub_block_ctl, address, self.num_str(length), lengths, comment)).rstrip())
        else:
            self.write_line(('%s %s %s' % (sub_block_ctl, address, comment)).rstrip())
